#ifndef DYN_SYST_TYPES
#define DYN_SYST_TYPES
#include <gsl/gsl_odeiv.h>
#include <gsl/gsl_roots.h>
/************************************************
 * types definitions used in the DS  analysis	*
 * package										*
 * (C) Alessandro Colombo, last update 15/12/10 *
 *                                              *
 ************************************************/


/* function pointers */
typedef double (*EVENT_FUNC_P) (const double time, const double state[], const int event_index, const void *pars);	/*pointer to the event function, use only whith event detection */
typedef int (*RHS_FUNC_P) (double time, const double state[], double rhs[], void *pars);	/* pointer to the rhs function f(x) */
typedef int (*JAC_FUNC_P) (double time, const double state[], double *dfdy, double dfdt[], void *pars);	/* pointer to the jacobian of f(x), used in ODEs whith implicit gsl integration methods */

/* system definition structure */
typedef struct
{
  RHS_FUNC_P rhsfun;		/*pointer to the function defining the right hand side */
  JAC_FUNC_P jacfun;		/*pointer to the function defining the jacobian */
  RHS_FUNC_P rhs_lin_part;  /*linear part of the right hand side (R^2n->R^2n)*/
  double t0;			/*initial time */
  double tf;			/*final time */
  double *x0;			/*initial state */
  void *pars;			/*parameters vector */
  double dt;			/*integration time step, used only for continuous-time systems. Set to 0.0 for adaptive time step. */
  int sysdim;			/*number of system variables */
  const gsl_odeiv_step_type *int_method;	/*integration method used by gsl routines */
  enum SYSTEM_TYPE
  { discrete, continuous } system_type;	/*defines continuous- or discrete-time system (ignored by ODE solver, which assumes continuous-time) */
} SYS_DEF;

/* event definition structure */
typedef struct
{
  EVENT_FUNC_P eventfun;	/*pointer to the event functions */
  const gsl_root_fsolver_type *event_method;	/*root-finding method used for the event detection */
  int n_eventfun;		/*number of event functions */
  int *direction;		/*vector indicating the direction of zero crossing needed to trigger an event  (0 for any direction) */
} EVENT_DEF;

/* integrator tolerances */
typedef struct
{
  double rel_tol;
  double abs_tol;
  double max_step_size; /* set =0 if not used */
  double min_step_size; /* set =0 if not used */
} INT_TOL;

/* Lyapunov exponents estimation tolerances */
typedef struct
{
  double lyap_tol;
  double min_time; /* minimum integration time before returning */
  double max_ratio;
  double min_ratio;
  double test_interval;
  double rescaling_factor; /* The norm of the initial linearised vectors is 1/(rescaling factor).
                              Large numbers can improve the estimation of the jacobian, however,
                              vectors should not be made smaller than the current state of the
                              nonlinear system times 1e-16.  If <=0 it is set to 1.0 */
} LYAP_TOL;

/* event detection tolerances */
typedef struct
{
  double rel_tol;
  double abs_tol;
  int max_iter;
} EVENT_TOL;

/* options */
typedef struct
{
  int *positive;		/* indices of positive variables */
  int n_positive;		/* number of positive variables */
  enum LYAPEXP_METHOD
  { QR, GS } lyapexp_method;	/* lyapunov exponents through QR factorization, or Gram Schmidt orthonormalization */
} OPTIONS;

#endif
