/************************************************
 * Lyapunov exponent estimation source file      *
 *                                               *
 * (C) Alessandro Colombo, last update 07/05/12  *
 *												*
 ************************************************/

/* inclusions */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_odeiv.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_errno.h>
#include "lyap_exp_1.1.h"
#include "small_functions_1.1.h"


#define MAX_COUNT_RENORM 10000


int lyap_exp (const SYS_DEF * sys_def, const INT_TOL * int_tol,
		  const LYAP_TOL * lyap_tol, const int n_exp, double *l_exp,
		  double *t_out, double state_out[], const OPTIONS * options)
{
	
	/* variables declaration */
	
	/* integrator variables */
	
	const gsl_odeiv_step_type *step_type;
	gsl_odeiv_step    *step;
	gsl_odeiv_control *ctrl;
	gsl_odeiv_evolve  *evol;
	gsl_odeiv_system  sys;
	
	/* counters and auxiliary variables */
	
	int i, 
	j, 
	count_renorm,
	count_renorm_max;
	enum CONVERGENCE
	{ not_yet, converged } convergence;
	int status;
	
	double rescaling_factor;
	
	/* QR factorization and linear system's state */
	
	double  step_size, /* integration step size */
	current_time,
	last_l_update_time,
	old_time,
	delta_time,
	state_interm_a[sys_def->sysdim],
	state_interm_b[sys_def->sysdim],
	f_val_a[sys_def->sysdim],
	f_val_b[sys_def->sysdim],
	d_lexp,
	last_lexp_tested[n_exp],
	convergence_par,
	last_test_time,
	max_ratio,
	min_ratio;
	
	double  state[sys_def->sysdim],	/* integrator state */
	old_state[sys_def->sysdim];
	
	double  extended_state[2*sys_def->sysdim], /* extended system variables */
	extended_state_err[2*sys_def->sysdim];
	gsl_odeiv_system  extended_sys;
	gsl_odeiv_step *extended_step;
	
	gsl_matrix *Q;		/* QR decomposition matrices */
	gsl_vector *Q_column;
	gsl_vector *qr_vec;
	gsl_vector *gs_vec;
	gsl_matrix *Q_tmp, *R_tmp;
	
#ifdef DEBUG
	if(sys_def->rhs_lin_part==NULL)
		printf
		("I am using the jacobian-free algorithm to integrate the variational equation\n");
	else
		printf
        ("I am using the user-defined linear system to integrate the variational equation \n");
	if(options->lyapexp_method==QR)
		printf
		("I am using QR orthonormalization \n");
	else if(options->lyapexp_method==GS)
		printf
		("I am using Gram Schmidt orthonormalization \n");
	else
		printf("options->lyapexp_metod is badly defined, I will use Gram Schmidt orthonormalization \n");
#endif
	
	/* check input data */
	
	if(n_exp > sys_def->sysdim)
    {
		fprintf (stderr,
				 "error: the system is %d dimensional but %d exponents were requested\n",sys_def->sysdim, 
				 n_exp);
		return GSL_EINVAL;
    }
	if (l_exp == NULL)
    {
		fprintf (stderr,
				 "error: vector l_exp is NULL, while I need a vector of %d elements to return the Lyapunov exponents\n",
				 n_exp);
		return GSL_EINVAL;
    }
	if (sys_def->rhsfun == NULL)
    {
		fprintf (stderr, "error: pointer to system's function is NULL\n");
		return GSL_EINVAL;
    }
	//  if ((sys_def->jacfun == NULL)&&((options->use_jacobian==all)||(options->use_jacobian==exponents)))
	//    {
	//      fprintf (stderr, "error: pointer to system's jacobian is NULL\n");
	//      return GSL_FAIL;
	//    }
	if (lyap_tol->rescaling_factor<=0.0)
		rescaling_factor=1.0;
	else
		rescaling_factor=lyap_tol->rescaling_factor;
	
	/* allocate matrices and vectors */
	
#ifdef DEBUG
	printf ("allocating matrices and vectors\n");
#endif
	
	Q = gsl_matrix_alloc (sys_def->sysdim, n_exp);
	Q_column = gsl_vector_alloc (sys_def->sysdim);
	Q_tmp = gsl_matrix_alloc (sys_def->sysdim, sys_def->sysdim);
	R_tmp = gsl_matrix_alloc (sys_def->sysdim, n_exp);
	qr_vec = gsl_vector_alloc (n_exp);
	gs_vec = gsl_vector_alloc (n_exp);
	
	/* integrators setup */
#ifdef DEBUG
	printf ("integrator and initial conditions setup\n");
#endif
	
	setup_integrator_and_initcond(sys_def, int_tol, &sys, &evol, &ctrl, &step, &step_type, &step_size, 
								  state, &current_time);
	
	
	extended_sys.function = sys_def->rhs_lin_part;
	extended_sys.jacobian = NULL;
	extended_sys.dimension = 2*sys_def->sysdim;
	extended_sys.params = sys_def->pars;
	
	extended_step = gsl_odeiv_step_alloc (step_type, 2*sys_def->sysdim);
	
	
	/* set initial vectors of the linearised system to a random value */
	
	srandom (time (NULL));
	for (i = 0; i < n_exp; i++)
		for (j = 0; j < sys_def->sysdim; j++) 
			gsl_matrix_set (Q, j, i, (random () % 1000) / 1000.0);
	
	/* change Q into an orthonormal matrix */
	gsl_linalg_QR_decomp (Q, qr_vec);
	gsl_linalg_QR_unpack (Q, qr_vec, Q_tmp, R_tmp);
	
	for (i = 0; i < n_exp; i++)
	{
		gsl_matrix_get_col (Q_column, Q_tmp, i);
		gsl_matrix_set_col (Q, i, Q_column);
	}
	/* rescale matrix Q by 1/rescaling_factor */
	gsl_matrix_scale(Q,1.0/rescaling_factor);
	
	
	/* set auxiliary variables */
	
	t_out[0] = sys_def->t0;
	for (i = 0; i < n_exp; i++) l_exp[i] = 0;
	
	d_lexp = 10 * lyap_tol->lyap_tol;
	
	convergence = not_yet;
	
	count_renorm = 0;
	count_renorm_max = 1;
	last_l_update_time = sys_def->t0;
	last_test_time = sys_def->t0;
	for (i = 0; i < n_exp; i++)
		last_lexp_tested[i]=0.0;
	
	/* start integration */
	
#ifdef DEBUG
	printf ("start integration\n");
#endif
	
	status = gsl_odeiv_evolve_reset (evol);
	if (status != GSL_SUCCESS)
    {
		fprintf (stderr,
				 "gsl_error: gsl_odeiv_evolve_reset returned %d\n", status);
		return status;
    }
	
	while ((current_time < sys_def->tf)
		   && ((convergence == not_yet) || current_time < sys_def->t0 + lyap_tol->min_time))
    {
		
		/* integration step */
		status=integration_step( sys_def, int_tol, sys, evol, ctrl,  step, &step_size, state, old_state, &current_time, &old_time, sys_def->tf);		
		if(status != GSL_SUCCESS){
			return status;
		}
		
		/* positive variables check */
		status=positive_variables_check(sys_def, options, sys, evol, step, &step_size, state, old_state, &current_time, &old_time);
		if(status != GSL_SUCCESS){
			return status;
		}	
        
        /* NAN check */ 
		if (state[0]!=state[0]){
			fprintf(stderr,"error: NAN\n");
			return -1;
		}
		
		/* if step was succesfull, integrate the linearised system with the same time-step */
		delta_time = current_time - old_time;
		if(sys_def->rhs_lin_part!=NULL)
			for (i = 0; i < n_exp; i++)
			{
				if (memcpy
					(extended_state, old_state,
					 sys_def->sysdim * sizeof (double)) == NULL)
				{
					fprintf (stderr,
							 "error: unable to copy the state vector onto the extended state vector\n");
					return GSL_ENOMEM;
				}
				for(j = 0; j <sys_def->sysdim; j++)
				{
					extended_state[j+sys_def->sysdim]=gsl_matrix_get (Q,j,i);
				}
				status =
				gsl_odeiv_step_apply (extended_step,
									  old_time,
									  delta_time,
									  extended_state, extended_state_err, NULL, NULL, &extended_sys);
				if (status != GSL_SUCCESS)
				{
					fprintf (stderr,
							 "gsl_error: gsl_odeiv_step_apply returned %d\n",
							 status);
					return status;
				}
				for(j = 0; j <sys_def->sysdim; j++)
				{
					gsl_matrix_set (Q,j,i,extended_state[j+sys_def->sysdim]);
				}
			}
		else
			for (i = 0; i < n_exp; i++)
			{
				
				/* intermediate step */
				
				sys_def->rhsfun (old_time, old_state, f_val_b, sys_def->pars);
				for (j = 0; j < sys_def->sysdim; j++)
				{
					state_interm_a[j] =
					old_state[j] + gsl_matrix_get (Q, j, i) * delta_time / 2.0;
				}
				//sys_def->rhsfun (old_time + delta_time / 2.0,
				//		   state_interm_a, f_val_a, sys_def->pars);
				sys_def->rhsfun (old_time,
								 state_interm_a, f_val_a, sys_def->pars);
				for (j = 0; j < sys_def->sysdim; j++)
				{
					gsl_vector_set (Q_column, j,
									gsl_matrix_get (Q, j, i) 
									+ f_val_a[j] - f_val_b[j]);
				}
				
				/* full step */
				
				for (j = 0; j < sys_def->sysdim; j++)
				{
					state_interm_a[j] =
					(old_state[j] + state[j]) / 2.0 +
					gsl_vector_get (Q_column, j) * delta_time;
					state_interm_b[j] =
					(old_state[j] + state[j]) / 2.0 - gsl_vector_get (Q_column, j) *
					delta_time;
				}
				sys_def->rhsfun (old_time + delta_time / 2.0,
								 state_interm_a, f_val_a, sys_def->pars);
				sys_def->rhsfun (old_time + delta_time / 2.0,
								 state_interm_b, f_val_b, sys_def->pars);
				
				for (j = 0; j < sys_def->sysdim; j++)
				{
					gsl_matrix_set (Q, j, i,
									gsl_matrix_get (Q, j, i) +
									(f_val_a[j] - f_val_b[j]) / 2.0);
				}
			}
		
		/* Update Lyapunov exponents*/
		count_renorm++;
		if((count_renorm>=count_renorm_max)||(current_time>=sys_def->tf))
		{
			/* rescale matrix Q by rescaling_factor */
			if(rescaling_factor!=1.0)
				gsl_matrix_scale(Q,rescaling_factor);
			
			/*  QR decomposition */
			if (options->lyapexp_method==QR)
			{
				gsl_linalg_QR_decomp (Q, qr_vec);
				gsl_linalg_QR_unpack (Q, qr_vec, Q_tmp, R_tmp);
				
				for (i = 0; i < n_exp; i++)
				{
					gsl_matrix_get_col (Q_column, Q_tmp, i);
					gsl_matrix_set_col (Q, i, Q_column);
					if(current_time > sys_def->t0)
						d_lexp =
						(-(current_time-last_l_update_time) * l_exp[i] +
						 log (fabs (gsl_matrix_get (R_tmp, i, i)))) /
						(current_time - sys_def->t0);
					else
						d_lexp = 0.0;
					l_exp[i] += d_lexp;
				}
				max_ratio=fabs(gsl_matrix_get (R_tmp, 0, 0));
				min_ratio=fabs(gsl_matrix_get (R_tmp, n_exp-1, n_exp-1));
			}
			
			/* else assume Gram Schmidt */
			else
			{
				gsl_matrix_gram_schmidt(Q, gs_vec);
				for (i = 0; i < n_exp; i++)
				{
					
					if(current_time > sys_def->t0)
						d_lexp = 
						(-(current_time-last_l_update_time) * l_exp[i] +
						 log (fabs (gsl_vector_get (gs_vec,i)))) /
						(current_time - sys_def->t0);
					else
						d_lexp = 0.0;
					l_exp[i] += d_lexp;
				}
				max_ratio=fabs(gsl_vector_get (gs_vec,0));
				min_ratio=fabs(gsl_vector_get (gs_vec,n_exp-1));
			}
			
			/* rescale matrix Q by 1/rescaling_factor */
			if(lyap_tol->rescaling_factor!=1.0)
				gsl_matrix_scale(Q,1.0/rescaling_factor);
			
			count_renorm=0;
			//printf("max_ratio=%e\n",max_ratio);
			//printf("min_ratio=%e\n",min_ratio);
			if((max_ratio<lyap_tol->max_ratio)&&(min_ratio>lyap_tol->min_ratio)&&(count_renorm_max<MAX_COUNT_RENORM))
				count_renorm_max+=1;
			else if(count_renorm_max>=2)
				count_renorm_max-=1;
			last_l_update_time=current_time;
			
			/* convergence check */
			if (current_time-last_test_time >= lyap_tol->test_interval){
				last_test_time=current_time;
				convergence_par = 0;
				for (i = 0; i < n_exp; i++){
					convergence_par = GSL_MAX(fabs ((l_exp[i]-last_lexp_tested[i]) / l_exp[i]),
											  convergence_par);
					last_lexp_tested[i]=l_exp[i];
				}
				if (convergence_par < lyap_tol->lyap_tol){
					convergence = converged;
				}
				else
					convergence = not_yet;
			}
		}
		if (!evol->last_step)
			break;
    }
	
	t_out[0] = current_time;
	for (i = 0; i < sys_def->sysdim; i++)
		state_out[i] = state[i];
	
#ifdef DEBUG
	printf ("done!\n");
#endif
	
	/* free memory */
	
	gsl_matrix_free (Q);
	gsl_matrix_free (Q_tmp);
	gsl_matrix_free (R_tmp);
	gsl_vector_free (Q_column);
	gsl_vector_free (qr_vec);
	
	gsl_odeiv_step_free (step);
	gsl_odeiv_evolve_free (evol);
	gsl_odeiv_control_free (ctrl);
	
	gsl_odeiv_step_free (extended_step);
	
	return GSL_SUCCESS;
}

