/************************************************
 * ODE solver source file                       *
 *                                              *
 * (C) Alessandro Colombo, last update 15/12/10 *
 *												*
 ************************************************/

/* 
 to do: 
 
 add min_step
 */

/* inclusions */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>  
#include <string.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_odeiv.h>
#include <gsl/gsl_roots.h>
#include "ode_solver_1.1.h"
#include "small_functions_1.1.h"

/* OS specific includes */
#ifdef __APPLE__
/* if it is OsX */
#include <malloc/malloc.h>
#else
/* if it is Linux */
#include <malloc.h>
#endif





/*main program*/

/*------------------------------------------------------------------------------------------------------------------------*/

int ode_solver(const SYS_DEF *sys_def, const INT_TOL *int_tol, double **t_out, double ***state_out, unsigned long int *n_el_out, const OPTIONS *options){
	
	int status;
	const gsl_odeiv_step_type *step_type;
	gsl_odeiv_step *step;
	gsl_odeiv_control *ctrl;
	gsl_odeiv_evolve *evol;
	gsl_odeiv_system sys;
	
	int i;
	
	double step_size,         	                  /* integration step size */
	current_time,
	next_out_time,                                  /* next output time for fixed step output */
 	old_time;
	double 	state[sys_def->sysdim],
	old_state[sys_def->sysdim];
	unsigned long int size_vec, j;		/*size of the memory allocated for output data*/
	
	
	/* integrators and initial conditions setup */
	status=setup_integrator_and_initcond(sys_def, int_tol,  &sys, &evol, &ctrl, &step, &step_type, &step_size, 
										 state, &current_time);
	if(status != GSL_SUCCESS){
		return status;
	}							
	
	/* output vectors allocation */
	status=output_vectors_allocation(sys_def, int_tol, &size_vec, n_el_out, state_out, state, t_out);
	if(status != GSL_SUCCESS){
		return status;
	}
	
	/* start integration*/
#ifdef DEBUG
	printf("start integration\n");
#endif
	
	status=gsl_odeiv_evolve_reset(evol);
	if(status != GSL_SUCCESS){
		fprintf(stderr,"gsl_error: gsl_odeiv_evolve_reset returned %d\n",status);
		return status;
	}
	
	/*if dt=0 then output each step*/
	if (sys_def->dt==0.0){
#ifdef DEBUG
		printf("variable time step output\n");
#endif
		while(current_time< sys_def->tf){
			
			/* integration step */
			status=integration_step( sys_def, int_tol, sys, evol, ctrl,  step, &step_size, state, old_state, &current_time, &old_time, sys_def->tf);		
			if(status != GSL_SUCCESS){
				return status;
			}
			
			/* positive variables check */
			status=positive_variables_check(sys_def, options, sys, evol, step, &step_size, state, old_state, &current_time, &old_time);
			if(status != GSL_SUCCESS){
				return status;
			}	
			
			if(!evol->last_step)
				break;
			
			/* NAN check */ 
			if (state[0]!=state[0]){
				fprintf(stderr,"Error: NAN\n");
				return GSL_EFAILED;
			}
			
			(*n_el_out)++;
			
			/*enlarge the output vector if needed*/
			status=resize_output(sys_def, &current_time, &size_vec, n_el_out, state_out, t_out);
			if(status != GSL_SUCCESS){
				return status;
			}	
			
			
			(*t_out)[(*n_el_out)-1]=current_time;
			for (i=0;i<sys_def->sysdim;i++)
				(*state_out)[i][(*n_el_out)-1]=state[i];
		}
		/* resize output vectors to the exact length */
		shrink_output(sys_def, *n_el_out, state_out, t_out);
	}
	
	/*if dt != 0 then output with fixed time step*/
	else{
#ifdef DEBUG
		printf("fixed time step output, dt = %f\n",sys_def->dt);
#endif
		*n_el_out=size_vec;
		for(j=1;j<size_vec;j++){
#ifdef DEBUG
			printf("step %ld of %ld\n",j,size_vec-1);
#endif
			if((sys_def->t0+(double)(j)*sys_def->dt)<sys_def->tf)
				next_out_time=(sys_def->t0+(double)(j)*sys_def->dt);
			else
				next_out_time=sys_def->tf;
			while(current_time < next_out_time){
				
				/* integration step */
				status=integration_step( sys_def, int_tol, sys, evol, ctrl,  step, &step_size, state, old_state, &current_time, &old_time, next_out_time);
				if(status != GSL_SUCCESS){
					return status;
				}	
				
				
				/* positive variables check */
				status=positive_variables_check(sys_def, options, sys, evol, step, &step_size, state, old_state, &current_time, &old_time);
				if(status != GSL_SUCCESS){
					return status;
				}	
				
				if(!evol->last_step){
					break;
				}
				
				/* NAN check */ 
				if (state[0]!=state[0]){
					fprintf(stderr,"error: NAN\n");
					return -1;
				}				
			}
			(*t_out)[j]=current_time;
			for (i=0;i<sys_def->sysdim;i++){
				(*state_out)[i][j]=state[i];
			}
			
			step_size=GSL_MIN(sys_def->dt,int_tol->max_step_size);
		}
		
	}
#ifdef DEBUG
	printf("done!\n");
#endif
	/* free memory*/
	gsl_odeiv_step_free(step);
	gsl_odeiv_evolve_free(evol);
	gsl_odeiv_control_free(ctrl);
	return GSL_SUCCESS;
}

/*------------------------------------------------------------------------------------------------------------------------*/

int ode_solver_noplot(const SYS_DEF *sys_def, const INT_TOL *int_tol, double *t_out, double state_out[], const OPTIONS *options){
	
	int status;
	
	const gsl_odeiv_step_type *step_type;
	gsl_odeiv_step *step;
	gsl_odeiv_control *ctrl;
	gsl_odeiv_evolve *evol;
	gsl_odeiv_system sys;
	
	int i;
	
	double 	step_size,	                  /* integration step size */
	current_time,
	old_time;
	double 	state[sys_def->sysdim],
	old_state[sys_def->sysdim];
	
	/* integrators and initial conditions setup */
	status=setup_integrator_and_initcond(sys_def, int_tol,  &sys, &evol, &ctrl, &step, &step_type, &step_size, 
										 state, &current_time);
	if(status != GSL_SUCCESS){
		return status;
	}	
	
	t_out[0]=sys_def->t0;
	
	
	/* start integration*/
#ifdef DEBUG
	printf("start integration\n");
#endif
	
	status=gsl_odeiv_evolve_reset(evol);
	if(status != GSL_SUCCESS){
		fprintf(stderr,"gsl_error: gsl_odeiv_evolve_reset returned %d\n",status);
		return status;
	}
	while(current_time<sys_def->tf){
		/* integration step */
		status=integration_step( sys_def, int_tol, sys, evol, ctrl,  step, &step_size, state, old_state, &current_time, &old_time, sys_def->tf);		
		if(status != GSL_SUCCESS){
			return status;
		}	
		
		/* positive variables check */
		status=positive_variables_check(sys_def, options, sys, evol, step, &step_size, state, old_state, &current_time, &old_time);
		if(status != GSL_SUCCESS){
			return status;
		}	
		
		if(!evol->last_step)
			break;
		
		/* NAN check */ 
		if (state[0]!=state[0]){
			fprintf(stderr,"error: NAN\n");
			return -1;
		}
	}
	
	t_out[0] = current_time;
	for(i=0;i<sys_def->sysdim;i++)
		state_out[i]=state[i];
#ifdef DEBUG
	printf("done!\n");
#endif
	/* free memory*/
	gsl_odeiv_step_free(step);
	gsl_odeiv_evolve_free(evol);
	gsl_odeiv_control_free(ctrl);
	return GSL_SUCCESS;
}

/*------------------------------------------------------------------------------------------------------------------------*/

int ode_solver_events(const SYS_DEF *sys_def, const INT_TOL *int_tol, const EVENT_DEF *event_def, const EVENT_TOL *event_tol, double **t_out, double ***state_out, unsigned long int *n_el_out, int *event_detected, int *convergence, const OPTIONS *options){
	
	/* integration variables and parameters */
	int status;
	const gsl_odeiv_step_type *step_type;
	gsl_odeiv_step *step;
	gsl_odeiv_control *ctrl;
	gsl_odeiv_evolve *evol;
	gsl_odeiv_system sys;
	
	int i;
	
	double step_size,         	                  /* integration step size */
	current_time,
 	next_out_time;					/* next output time for fixed step output */
	double state[sys_def->sysdim];
	unsigned long int size_vec=0, j;		/*size of the memory allocated for output data*/
	/* end integration variables and parameters */
	
	/* event detection variables and parameters */
	int last_sign[event_def->n_eventfun];
	EVENT_WRAPPER_PARS ewpars;
	/* end event detection variables and parameters */
	
	/* integrators and initial conditions setup */
	status=setup_integrator_and_initcond(sys_def, int_tol,  &sys, &evol, &ctrl, &step, &step_type, &step_size, 
										 state, &current_time);
	if(status != GSL_SUCCESS){
		return status;
	}	
	
	/* output vectors allocation */
	status=output_vectors_allocation(sys_def, int_tol, &size_vec, n_el_out, state_out, state, t_out);
	if(status != GSL_SUCCESS){
		return status;
	}	
	
	/* event setup */
	status=setup_event(sys_def, &ewpars, event_def, last_sign, event_detected, convergence);
	if(status != GSL_SUCCESS){
		return status;
	}	
	
	/* start integration*/
#ifdef DEBUG
	printf("start integration\n");
#endif
	
	status=gsl_odeiv_evolve_reset(evol);
	if(status != GSL_SUCCESS){
		fprintf(stderr,"gsl_error: gsl_odeiv_evolve_reset returned %d\n",status);
		return status;
	}
	
	/*if dt=0 then output each step*/
	if (sys_def->dt==0.0){
#ifdef DEBUG
		printf("variable time step output\n");
#endif
		while(current_time< sys_def->tf){
			
			/* integration step */
			status=integration_step( sys_def, int_tol, sys, evol, ctrl,  step, &step_size, state, ewpars.state, &current_time, &(ewpars.time), sys_def->tf);		
			if(status != GSL_SUCCESS){
				return status;
			}
			
			/* positive variables check */
			status=positive_variables_check( sys_def, options, sys, evol, step, &step_size, state, ewpars.state, &current_time, &(ewpars.time));
			if(status != GSL_SUCCESS){
				return status;
			}	
			
			/* NAN check */ 
			if (state[0]!=state[0]){
				fprintf(stderr,"error: NAN\n");
				return -1;
			}
			
			(*n_el_out)++;
			
			/*enlarge the output vector if needed*/
			status=resize_output(sys_def, &current_time, &size_vec, n_el_out, state_out, t_out);
			if(status != GSL_SUCCESS){
				return status;
			}	
			
			/* event check*/
			status=event_check( sys_def, &ewpars, event_def, event_tol, &sys, step, last_sign, state, &current_time, event_detected, convergence);
			if(status != GSL_SUCCESS){
				return status;
			}	
			
			(*t_out)[(*n_el_out)-1]=current_time;
			for (i=0;i<sys_def->sysdim;i++)
				(*state_out)[i][(*n_el_out)-1]=state[i];
			
			if (*event_detected>=0)
				break;
			
			if(!evol->last_step)
				break;
		}
		
		/* resize output vectors to the exact length */
		shrink_output(sys_def, *n_el_out, state_out, t_out);
	}
	
	/*if dt != 0 then output with fixed time step*/
	else{
#ifdef DEBUG
		printf("fixed time step output, dt = %f\n",sys_def->dt);
#endif
		*n_el_out=0;
		for(j=1;j<size_vec;j++){
#ifdef DEBUG
			printf("step %ld\n",j);
#endif		
			if((sys_def->t0+(double)(j)*sys_def->dt)<sys_def->tf)
				next_out_time=(sys_def->t0+(double)(j)*sys_def->dt);
			else
				next_out_time=sys_def->tf;
			while(current_time <= (next_out_time) ){
				
				/* integration step */
				status=integration_step( sys_def, int_tol, sys, evol, ctrl,  step, &step_size, state, ewpars.state, &current_time, &(ewpars.time), next_out_time);		
				if(status != GSL_SUCCESS){
					return status;
				}
				
				/* positive variables check */
				status=positive_variables_check( sys_def, options, sys, evol, step, &step_size, state, ewpars.state, &current_time, &(ewpars.time));
				if(status != GSL_SUCCESS){
					return status;
				}	
				
				/* NAN check */ 
				if (state[0]!=state[0]){
					fprintf(stderr,"error: NAN\n");
					return -1;
				}
				
				/* event check*/
				status=event_check( sys_def, &ewpars, event_def, event_tol, &sys, step, last_sign, state, &current_time, event_detected, convergence);
				if(status != GSL_SUCCESS){
					return status;
				}	
				
				if (*event_detected>=0)
					break;
				
				if(!evol->last_step)
					break;
			}
			(*t_out)[j]=current_time;
			for (i=0;i<sys_def->sysdim;i++)
				(*state_out)[i][j]=state[i];
			
			step_size=sys_def->dt;
			if (*event_detected>=0)
			{
				*n_el_out=j+1;
				break;
			}
		}
		
		if (*event_detected<0)
			*n_el_out=size_vec;
		/* resize output vectors to the exact length */
		shrink_output(sys_def, *n_el_out, state_out, t_out);
		
	}
#ifdef DEBUG
	printf("done!\n");
#endif
	/* free memory*/
	free(ewpars.state);
	gsl_odeiv_step_free(step);
	gsl_odeiv_evolve_free(evol);
	gsl_odeiv_control_free(ctrl);
	return GSL_SUCCESS;;
}

/*------------------------------------------------------------------------------------------------------------------------*/

int ode_solver_events_noplot(const SYS_DEF *sys_def, const INT_TOL *int_tol, const EVENT_DEF *event_def, const EVENT_TOL *event_tol, double *t_out, double state_out[], int *event_detected, int *convergence, const OPTIONS *options){
	
	/* integration variables and parameters           */
	int status;	
	
	const gsl_odeiv_step_type *step_type;
	gsl_odeiv_step *step;
	gsl_odeiv_control *ctrl;
	gsl_odeiv_evolve *evol;
	gsl_odeiv_system sys;
	
	int i;
	
	double step_size,         /* integration step size */
	current_time;
	double state[sys_def->sysdim];
	/* end integration variables and parameters */	
	
	/* event detection variables and parameters     */
	int last_sign[event_def->n_eventfun];
	EVENT_WRAPPER_PARS ewpars;
	/* end event detection variables and parameters */
	
	/* integrators and initial conditions setup */
	status=setup_integrator_and_initcond(sys_def, int_tol,  &sys, &evol, &ctrl, &step, &step_type, &step_size, 
										 state, &current_time);
	if(status != GSL_SUCCESS){
		return status;
	}	
	
	/* event setup */
	status=setup_event(sys_def, &ewpars, event_def, last_sign, event_detected, convergence);
	if(status != GSL_SUCCESS){
		return status;
	}	
	
	t_out[0]=sys_def->t0;
	
	/* start integration*/
#ifdef DEBUG
	printf("start integration\n");
#endif
	
	status=gsl_odeiv_evolve_reset(evol);
	if(status != GSL_SUCCESS){
		fprintf(stderr,"gsl_error: gsl_odeiv_evolve_reset returned %d\n",status);
		return status;
	}
	while(current_time<sys_def->tf){
		/* integration step */
		status=integration_step( sys_def, int_tol, sys, evol, ctrl,  step, &step_size, state, ewpars.state, &current_time, &(ewpars.time), sys_def->tf);		
		if(status != GSL_SUCCESS){
			return status;
		}
		
		/* positive variables check */
		status=positive_variables_check( sys_def, options, sys, evol, step, &step_size, state, ewpars.state, &current_time, &(ewpars.time));
		if(status != GSL_SUCCESS){
			return status;
		}	
		
		/* NAN check */ 
		if (state[0]!=state[0]){
			fprintf(stderr,"error: NAN\n");
			return -1;
		}
		
		/* event check*/
		status=event_check( sys_def, &ewpars, event_def, event_tol, &sys, step, last_sign, state, &current_time, event_detected, convergence);
		if(status != GSL_SUCCESS){
			return status;
		}	
		
		if (*event_detected>=0)
			break;
		if(!evol->last_step)
			break;
	}
	
	t_out[0] = current_time;
	for(i=0;i<sys_def->sysdim;i++)
		state_out[i]=state[i];
#ifdef DEBUG	
	printf("done!\n");
#endif
	/* free memory*/
	free(ewpars.state);
	gsl_odeiv_step_free(step);
	gsl_odeiv_evolve_free(evol);
	gsl_odeiv_control_free(ctrl);
	return GSL_SUCCESS;
}
