/************************************************
 * Functions library							*
 *                                              *
 * (C) Alessandro Colombo, last update 18/07/12 *
 *												*
 ************************************************/
 
 /* inclusions */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>  
#include <string.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_odeiv.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_linalg.h>
#include "small_functions_1.1.h"

/* OS specific includes */
#ifdef __APPLE__
/* if it is OsX */
#include <malloc/malloc.h>
#else
/* if it is Linux */
#include <malloc.h>
#endif

/*------------------------------------------------------------------------------------------------------------------------*/

inline double event_wrapper(const double t, void * ewpars){
	
	int status;
	int i;
	double state[((EVENT_WRAPPER_PARS*)ewpars)->sysdim],state_err[((EVENT_WRAPPER_PARS*)ewpars)->sysdim];
	double output;
	const gsl_odeiv_step_type *step_type;
	gsl_odeiv_step *step;
	gsl_odeiv_system sys;
	double (*eventfun)(const double current_time, const double state[], const int event_index, const void *pars);
	
	step_type=((EVENT_WRAPPER_PARS*)ewpars)->method;
	step=gsl_odeiv_step_alloc(step_type, ((EVENT_WRAPPER_PARS*)ewpars)->sysdim);
	sys.function=((EVENT_WRAPPER_PARS*)ewpars)->sys_function;
	sys.jacobian=((EVENT_WRAPPER_PARS*)ewpars)->jac_function;
	sys.dimension=((EVENT_WRAPPER_PARS*)ewpars)->sysdim;
	sys.params=((EVENT_WRAPPER_PARS*)ewpars)->pars;
	eventfun=((EVENT_WRAPPER_PARS*)ewpars)->ev_function;
	for(i=0;i<((EVENT_WRAPPER_PARS*)ewpars)->sysdim;i++)
		state[i]=((EVENT_WRAPPER_PARS*)ewpars)->state[i];
	
	status=gsl_odeiv_step_apply(step, ((EVENT_WRAPPER_PARS*)ewpars)->time, t, state, state_err, NULL, NULL, &sys);
	
	output = eventfun(((EVENT_WRAPPER_PARS*)ewpars)->time+t, state, ((EVENT_WRAPPER_PARS*)ewpars)->event_index, ((EVENT_WRAPPER_PARS*)ewpars)->pars);
	
	/* free memory*/
	gsl_odeiv_step_free(step);
	return output;
}


/*------------------------------------------------------------------------------------------------------------------------*/
inline int integration_step(const SYS_DEF *sys_def, const INT_TOL *int_tol, gsl_odeiv_system sys, 
							gsl_odeiv_evolve *evol, gsl_odeiv_control *ctrl, gsl_odeiv_step *step, 
							double *step_size, double *state, double * old_state,
							double * current_time, double * old_time, double next_out_time){
	double next_time;
	int status;
	
	if(memcpy(old_state, state, sys_def->sysdim*sizeof(double))==NULL){
		fprintf(stderr,"error: unable to copy the state vector onto the old_state vector\n");
		return GSL_ENOMEM;
	}
	*old_time=*current_time;
	
	if (int_tol->max_step_size > 0)
		next_time = GSL_MIN (next_out_time, *current_time + int_tol->max_step_size);
	else
		next_time = next_out_time;

	status=gsl_odeiv_evolve_apply(evol, ctrl, step, &sys, current_time, next_time, step_size, state);
	if(status != GSL_SUCCESS){
		fprintf(stderr,"gsl_error: gsl_odeiv_evolve_apply returned %d\n",status);
		return status;
	}
	
	if((*step_size < int_tol->min_step_size)&(*step_size<next_out_time-*current_time)){
		fprintf(stderr,"error: step size below min_step_size\n");
		return GSL_ENOPROG;
	}
	
	return GSL_SUCCESS;
}
/*------------------------------------------------------------------------------------------------------------------------*/
inline int positive_variables_check(const SYS_DEF *sys_def, const OPTIONS *options, const gsl_odeiv_system sys,
									gsl_odeiv_evolve *evol, gsl_odeiv_step *step, double * step_size, 
									double *state, double * old_state, double * current_time, const double * old_time){
	int i;
	int status;
	double state_err[sys_def->sysdim];
	
	if((options->n_positive>0)&&(options->positive!=NULL)){
		*step_size=*current_time-*old_time;
		for(i=0;i<options->n_positive;i++){
			while(state[options->positive[i]]<-0.0){
				i=0; /*reset the counter, in order to check again all the variables*/
				if(memcpy(state, old_state, sys_def->sysdim*sizeof(double))==NULL){
					fprintf(stderr,"error: unable to copy the old state vector onto the new one\n");
					return GSL_ENOMEM;
				}
				*current_time=*old_time;
				*step_size=(*step_size)*0.5;
				if( (*step_size) <=0.0){
					fprintf(stderr,"error: step size below 0 (%2.16f), unable to keep variable %d positive\n",*step_size,options->positive[i]);
					return GSL_ENOPROG;
				}
				status=gsl_odeiv_step_apply(step, *current_time, *step_size, state, state_err, NULL, NULL, &sys);
				if(status != GSL_SUCCESS){
					fprintf(stderr,"gsl_error: gsl_odeiv_step_apply returned %d\n",status);
					return status;
				}
				(*current_time)+=(*step_size);
				evol->last_step=*step_size;
			}
		}
	}
	else if((options->n_positive>0)&&(options->positive==NULL)){
		fprintf(stderr,"error: %d positive variables expected, but the indices vector is NULL\n",options->n_positive);
		return GSL_EINVAL;
	}
	
	return GSL_SUCCESS;
}

/*------------------------------------------------------------------------------------------------------------------------*/
inline int setup_integrator_and_initcond(const SYS_DEF *sys_def, const INT_TOL *int_tol, gsl_odeiv_system *sys, 
										 gsl_odeiv_evolve **evol, gsl_odeiv_control **ctrl, gsl_odeiv_step **step,
										 const gsl_odeiv_step_type **step_type, double * step_size, double *state, double * current_time){
#ifdef DEBUG
	printf("integrator setup\n");
#endif
	if (sys_def->tf<sys_def->t0){
		fprintf(stderr,"error: final integration time smaller than initial time.\nTo integrate backward in time, multiply f by -1\n");
		return GSL_EINVAL;
	}	
	
	if (sys_def->dt<0){
		fprintf(stderr,"error: dt<0\nTo integrate backward in time, multiply f by -1\n");
		return GSL_EINVAL;
	}

	*step_type=sys_def->int_method;
	*step=gsl_odeiv_step_alloc(*step_type, sys_def->sysdim);
	*ctrl=gsl_odeiv_control_y_new(int_tol->abs_tol, int_tol->rel_tol);
	*evol=gsl_odeiv_evolve_alloc(sys_def->sysdim);
	sys->function=sys_def->rhsfun;
	sys->jacobian=sys_def->jacfun;
	sys->dimension=sys_def->sysdim;
	sys->params=sys_def->pars;
	
#ifdef DEBUG
	printf("initial conditions setup\n");
#endif
	if(memcpy(state, sys_def->x0, sys_def->sysdim*sizeof(double))==NULL){
		fprintf(stderr,"error: unable to copy the inital conditions vector onto the state vector\n");
		return GSL_ENOMEM;
	}
	*current_time=sys_def->t0;
	if (sys_def->dt==0.0){
		if (int_tol->min_step_size>0){
			*step_size=int_tol->min_step_size;
#ifdef DEBUG
			printf("initial integration step set to min_step_size\n");
#endif
		}
		else{
			*step_size=int_tol->max_step_size;
#ifdef DEBUG
			printf("min_step_size = 0, initial integration step set to max_step_size\n");
#endif
		}
	}
	else{
		if (int_tol->min_step_size>0){
			*step_size=GSL_MIN(sys_def->dt,int_tol->min_step_size);
#ifdef DEBUG
			printf("initial integration step set to min_step_size\n");
#endif
		}
		else{
			*step_size=GSL_MIN(sys_def->dt,int_tol->max_step_size);
#ifdef DEBUG
			printf("min_step_size = 0, initial integration step set to max_step_size\n");
#endif
		}
	}

	return GSL_SUCCESS;
}

/*------------------------------------------------------------------------------------------------------------------------*/
inline int output_vectors_allocation(const SYS_DEF *sys_def, const INT_TOL *int_tol, unsigned long int * size_vec, 
									 unsigned long int *n_el_out, double *** state_out, double *state, double **t_out){
	int i;
	
#ifdef DEBUG
	printf("output vectors allocation\n");
#endif
	if (ceil((sys_def->tf-sys_def->t0)/int_tol->max_step_size) > LONG_MAX){
		fprintf(stderr,"Error (ode_solver), vector output too long\n");
		return GSL_EINVAL;
	}	
	
	(*size_vec)=ceil((sys_def->tf-sys_def->t0)/int_tol->max_step_size)+1;
#ifdef DEBUG
	printf("size_vec = %lu\n",*size_vec);
#endif
	*t_out = (double*)malloc((*size_vec)*sizeof(double));
	if (*t_out==NULL){
		fprintf(stderr,"Error (ode_solver), not enough memory to allocate time vector\n");
		return GSL_ENOMEM;
	}
	
	*state_out = malloc(sys_def->sysdim*sizeof(double*));
	if (*state_out==NULL){
		fprintf(stderr,"Error (ode_solver), not enough memory to allocate state vector\n");
		return GSL_ENOMEM;
	}
	
#ifdef DEBUG
	printf("state vector base allocated\n");
#endif
	for (i=0;i<sys_def->sysdim;i++){
#ifdef DEBUG
		printf("allocating space for state vector %d...",i);
#endif
		fflush(stdout);
		(*state_out)[i] = malloc((*size_vec)*sizeof(double));
		fflush(stdout);
		if ((*state_out)[i]==NULL){
			fprintf(stderr,"Error (ode_solver), not enough memory to allocate state vector\n");
			return GSL_ENOMEM;
		}
#ifdef DEBUG
		printf("done\n");
#endif
	}
#ifdef DEBUG
	printf("memory allocated\n");
#endif
	(*t_out)[0]=sys_def->t0;
	for (i=0;i<sys_def->sysdim;i++)
		(*state_out)[i][0]=state[i];
	*n_el_out=1;
	
	return GSL_SUCCESS;
}

/*------------------------------------------------------------------------------------------------------------------------*/
inline int setup_event(const SYS_DEF *sys_def, EVENT_WRAPPER_PARS *ewpars, const EVENT_DEF *event_def, int * last_sign, 
					   int *event_detected, int *convergence){
	
	int i;
	double event_value;
	
	/* event wrapper setup */
	ewpars->sys_function=sys_def->rhsfun;
	ewpars->jac_function=sys_def->jacfun;
	ewpars->ev_function=event_def->eventfun;
	ewpars->sysdim=sys_def->sysdim;
	ewpars->method=sys_def->int_method;
	ewpars->pars=sys_def->pars;
	ewpars->state=(double*)malloc(sys_def->sysdim*sizeof(double));
	
	*event_detected=-1;
	*convergence=0;
	
#ifdef DEBUG
	printf("determining initial sign of event functions\n");
#endif
	if (event_def->eventfun!=NULL){
		for (i=0;i<event_def->n_eventfun;i++){
			event_value=event_def->eventfun(sys_def->t0, sys_def->x0, i, sys_def->pars);
			last_sign[i]=GSL_SIGN(event_value);
#ifdef DEBUG
			printf("initial sign of event function %d is %d\n",i,last_sign[i]);
#endif	
		}
	}
	else{
		fprintf(stderr,"error: event function undefined\n");
		return GSL_EINVAL;
	}
	
	return GSL_SUCCESS;
}
/*------------------------------------------------------------------------------------------------------------------------*/
inline int event_check(const SYS_DEF * sys_def, EVENT_WRAPPER_PARS * ewpars, const EVENT_DEF *event_def, 
					   const EVENT_TOL *event_tol,  gsl_odeiv_system *sys, gsl_odeiv_step *step, int * last_sign,
					   double * state, double * current_time, int * event_detected, int * convergence){
	int current_sign[event_def->n_eventfun];
	
	int i,j,iter,status;
	double 	time_step=0,
	t_lo=0,
	t_hi=0, 
	root=0;
	double state_err[((EVENT_WRAPPER_PARS*)ewpars)->sysdim];
	double event_value;
	gsl_function F;
	const gsl_root_fsolver_type *fsType;
	gsl_root_fsolver *fsolver;
	
	time_step = *current_time-ewpars->time;
	for (i = 0;i<event_def->n_eventfun;i++){
		event_value=event_def->eventfun(*current_time, state, i, sys_def->pars);
		current_sign[i]=GSL_SIGN(event_value);
		if((last_sign[i] * current_sign[i] < 0) && (current_sign[i]*event_def->direction[i] >= 0)){
#ifdef DEBUG
			printf ("Event %d detected\n",i);
#endif
			ewpars->event_index = i;
			F.function = &event_wrapper;
			F.params = ewpars;
			fsType = event_def->event_method;
			fsolver = gsl_root_fsolver_alloc(fsType);
			t_lo = 0.0;
			t_hi = *current_time - ewpars->time;
			gsl_root_fsolver_set(fsolver,&F,t_lo,t_hi);
			iter = 0;
			do{
				iter++;
				status = gsl_root_fsolver_iterate (fsolver);
				root = gsl_root_fsolver_root (fsolver);
				t_lo = gsl_root_fsolver_x_lower (fsolver);
				t_hi = gsl_root_fsolver_x_upper (fsolver);
				status = gsl_root_test_interval (t_lo, t_hi, event_tol->abs_tol, event_tol->rel_tol);
#ifdef DEBUG
				if (status == GSL_SUCCESS)
					printf ("Converged in %d iterations\n",iter);
#endif
			}
			while (status == GSL_CONTINUE && iter < event_tol->max_iter);
			if (status == GSL_SUCCESS)
				*convergence=1;
#ifdef DEBUG
			else
				printf ("No convergence, try increasing the maximum number of iterations\n",iter);
#endif
			if(time_step>t_lo){
				time_step = t_lo;
				*event_detected = i;
			}
			gsl_root_fsolver_free(fsolver);
		}
		last_sign[i]=current_sign[i];
	}
	if (*event_detected>=0){
#ifdef DEBUG
		printf("event %d happens first\n",*event_detected);
#endif
		/*do a single step to get to the event point*/
		for(j=0;j<sys_def->sysdim;j++)
			state[j]=ewpars->state[j];
		status=gsl_odeiv_step_apply(step, ewpars->time, time_step, state, state_err, NULL, NULL, sys);
		if(status != GSL_SUCCESS){
			fprintf(stderr,"gsl_error: gsl_odeiv_step_apply returned %d\n",status);
			return status;
		}
		*current_time=ewpars->time+time_step;
	}
	return GSL_SUCCESS;
}
/*------------------------------------------------------------------------------------------------------------------------*/
inline int resize_output(const SYS_DEF * sys_def, double * current_time,
						 unsigned long int * size_vec, unsigned long int *n_el_out, double *** state_out, double **t_out){
	
	double d_size_vec;
	double av_step_size=0;
	int i;
	if ((*n_el_out)>=*size_vec){
		if(*current_time<sys_def->tf)
			av_step_size=fabs((*current_time)-sys_def->t0)/(*n_el_out-1);
		d_size_vec=ceil((sys_def->tf-*current_time)/av_step_size+1.0);
		(*size_vec)+=d_size_vec;
		(*t_out) = realloc(*t_out,(*size_vec)*sizeof(double));
		if (*t_out==NULL){
			fprintf(stderr,"Error (ode_solver), not enough memory to allocate the time vector\n");
			return GSL_ENOMEM;
		}
		for (i=0;i<sys_def->sysdim;i++){;
			(*state_out)[i] = realloc((*state_out)[i],(*size_vec)*sizeof(double));
			if ((*state_out)[i]==NULL){
				fprintf(stderr,"Error (ode_solver), not enough memory to allocate the state vector\n");
				return GSL_ENOMEM;
			}
		}
#ifdef DEBUG
		printf("vectors reallocated, new size_vec is %ld\n",size_vec);
#endif
	}
	else{
	
	}
	return GSL_SUCCESS;			
}
/*------------------------------------------------------------------------------------------------------------------------*/
inline int shrink_output(const SYS_DEF * sys_def, unsigned long int n_el_out, double *** state_out, double ** t_out){

	int i;
	
	*t_out = realloc(*t_out,n_el_out*sizeof(double));
	if (*t_out==NULL){
		fprintf(stderr,"Error (ode_solver), not enough memory to allocate time vector\n");
		return GSL_ENOMEM;
	}
	for (i=0;i<sys_def->sysdim;i++){
		(*state_out)[i] = realloc((*state_out)[i],n_el_out*sizeof(double));	
		if ((*state_out)[i]==NULL){
			fprintf(stderr,"Error (ode_solver), not enough memory to allocate state vector\n");
			return GSL_ENOMEM;
		}
	}
#ifdef DEBUG
	printf("output vectors reallocated, new size_vec is %ld\n",*n_el_out);
#endif
	return GSL_SUCCESS;
}
/*------------------------------------------------------------------------------------------------------------------------*/

inline int rhs_wrapper (double time, const double state[], double rhs[], void *Wpars)
{
	SYS_DEF *sys_def = Wpars;
	void *pars = sys_def->pars;
	JAC_FUNC_P jacfun = sys_def->jacfun;
	double jac_vect[sys_def->sysdim * sys_def->sysdim];
	double drhsdt[sys_def->sysdim];
	gsl_matrix_view jac_matrix;
	gsl_vector_view state_v_vector;
	gsl_vector_view rhs_v_vector;
	
	sys_def->rhsfun (time, state, rhs, pars);
	jacfun (time, state, jac_vect, drhsdt, pars);
	jac_matrix =
	gsl_matrix_view_array (jac_vect, sys_def->sysdim, sys_def->sysdim);
	state_v_vector =
	gsl_vector_view_array ((double*)&(state[sys_def->sysdim]),
						   sys_def->sysdim);
	rhs_v_vector =
	gsl_vector_view_array (&(rhs[sys_def->sysdim]),
						   sys_def->sysdim);
	gsl_blas_dgemv (CblasNoTrans, 1.0, &jac_matrix.matrix,
					&state_v_vector.vector, 0.0, &rhs_v_vector.vector);
	return 0;
}
/*------------------------------------------------------------------------------------------------------------------------*/

/* gsl_matrix_gram_schmidt: applied to A (square), norms stored in v */
inline void gsl_matrix_gram_schmidt(gsl_matrix *A, gsl_vector *v)
{
	int i,j;
	double scalar_product;
	double norm;
	
	/* allocate space for temporary vectors */
	gsl_vector *a = gsl_vector_alloc(A->size1);
	gsl_vector *b = gsl_vector_alloc(A->size1);
	gsl_vector *proj = gsl_vector_alloc(A->size1);
	
	if (A->size2 != v->size)
	{
		fprintf (stderr, "gsl_matrix_gram_schmidt error: matrix rows have different length than coefficients vector\n");
		return;
	}
	
	for(i=0; i<A->size2; i++)
	{
		/* extract a vector: a */
		gsl_matrix_get_col(a,A,i);
		/* normalize a */
		gsl_vector_normalize(a,&norm);
		gsl_vector_set(v,i,norm);
		
		for(j=i+1; j<A->size2; j++)
		{
			/* extract another vector: b */
			gsl_matrix_get_col(b,A,j);
			
			/* scalar product a.b */
			gsl_blas_ddot (a, b, &scalar_product);
			
			/* proj = s*a */
			gsl_vector_memcpy(proj,a);
			gsl_vector_scale(proj,scalar_product);
			
			/* b = b - a */
			gsl_vector_sub(b,proj);
			gsl_matrix_set_col(A,j,b);
		}
		
		/* set matrix*/
		gsl_matrix_set_col(A,i,a);
	}
	
	/* free */
	gsl_vector_free(a);
	gsl_vector_free(b);
	gsl_vector_free(proj);
	
	return;
}
/*------------------------------------------------------------------------------------------------------------------------*/

inline void gsl_vector_normalize(gsl_vector *a, double *norm)
{
	*norm = gsl_blas_dnrm2(a);
	gsl_blas_dscal(1.0/(*norm),a);
	return;
}