/************************************************
 * Main file									*
 *                                              *
 * (C) Alessandro Colombo, last update 15/12/10 *
 ************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include "ode_solver_1.1.h"
#include <gsl/gsl_math.h>
#include <gsl/gsl_odeiv.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_errno.h>

int rhsfun(const double time, const double state[2], double rhs[2], void *pars);
double eventfun(const double t, const double x[2], const int event_index, const void *pars);

int main(int argc, char **argv){
	
	double x0[2]={1,2};
	double pars[2]={-0.1,0.5};
	SYS_DEF sysdef;
	INT_TOL inttol;
	EVENT_DEF eventdef;
	EVENT_TOL eventtol;
	OPTIONS options;
	int flag,i,j;
	
	double *tout1;
	double **stateout1;
	unsigned long int nelout1;
	
	double tout2;
	double stateout2[2];
	
	double tout3;
	double stateout3[2];
	int eventdetected;
	int convergence;
	
	double *tout4;
	double **stateout4;
	unsigned long int nelout4;
	
	FILE* out_file;
	
	sysdef.rhsfun=rhsfun;
	sysdef.jacfun=NULL;
	sysdef.t0=0.0;
	sysdef.tf=20000.0;
	sysdef.x0=(double*)malloc(2*sizeof(double));
	sysdef.x0[0]=x0[0];
	sysdef.x0[1]=x0[1];
	sysdef.pars=pars;
	sysdef.dt=0.0;
	sysdef.sysdim=2;
	sysdef.int_method=gsl_odeiv_step_rkf45;
	
	inttol.rel_tol=1.0E-8;
	inttol.abs_tol=1.0E-8;
	inttol.max_step_size=1E-1;
	inttol.min_step_size=1E-2;
	
	eventdef.eventfun=eventfun;
	eventdef.event_method=gsl_root_fsolver_brent;
	eventdef.n_eventfun=1;
	eventdef.direction=(int*)malloc(sizeof(int));
	
	eventtol.rel_tol=0.0;
	eventtol.abs_tol=1.0E-8;
	eventtol.max_iter=10;
	
	options.n_positive=0;
	
	tout1=NULL;
	
	printf("\ntest ode_solver\n");
	flag=ode_solver(&sysdef, &inttol, &tout1, &stateout1, &nelout1, &options);
	if (flag==GSL_SUCCESS){
		out_file=fopen("out1.dat","w");
		for(i=0;i<nelout1;i++){
			fprintf(out_file,"%f\t",tout1[i]);
			fprintf(out_file,"%f\t",stateout1[0][i]);
			fprintf(out_file,"%f\n",stateout1[1][i]);
		}
		fclose(out_file);
		printf("done, data saved in out1.dat\n");
		free(tout1);
		for(i=0;i<2;i++)
			free(stateout1[i]);
		free(stateout1);
	}
	else printf("flag = %d\n", flag);
	
	printf("\ntest ode_solver_noplot\n");
	flag=ode_solver_noplot(&sysdef, &inttol, &tout2, stateout2, &options); 
	if (flag==GSL_SUCCESS){
		out_file=fopen("out2.dat","w");
		fprintf(out_file,"%f\t",tout2);
		fprintf(out_file,"%f\t",stateout2[0]);
		fprintf(out_file,"%f\n",stateout2[1]);
		fclose(out_file);
		printf("done, data saved in out2.dat\n");
	}
	else printf("flag = %d\n", flag);
	
	
	printf("\ntest ode_solver_events_noplot\n");
	eventdef.direction[0]=-1;
	sysdef.t0=0.0;
	tout3=0;
	out_file=fopen("out3.dat","w");
	
	for(i=0;i<4;i++){
		sysdef.t0=tout3;
		flag = ode_solver_events_noplot(&sysdef, &inttol, &eventdef, &eventtol, &tout3, stateout3, &eventdetected, &convergence, &options);
		if (flag==GSL_SUCCESS){
			if(eventdetected>=0){
				fprintf(out_file,"%f\t",tout3);
				fprintf(out_file,"%f\t",stateout3[0]);
				fprintf(out_file,"%f\n",stateout3[1]);
				printf("event %d detected, data saved in out3.dat\n",i);
				eventdef.direction[0]=-eventdef.direction[0];
			}
			sysdef.x0[0]=stateout3[0];
			sysdef.x0[1]=stateout3[1];
		}
		else
			printf("flag = %d\n", flag);
	}
	fclose(out_file);
	
	printf("\ntest ode_solver_events\n");
	eventdef.direction[0]=-1;
	sysdef.t0=0.0;
	sysdef.x0[0]=x0[0];
	sysdef.x0[1]=x0[1];
	out_file=fopen("out4.dat","w");
	for(i=0;i<4;i++){
		flag = ode_solver_events(&sysdef, &inttol, &eventdef, &eventtol, &tout4, &stateout4, &nelout4, &eventdetected, &convergence, &options);
		if (flag==GSL_SUCCESS){
			if(eventdetected>=0){
				for(j=0;j<nelout4;j++){
					fprintf(out_file,"%d\t",i);
					fprintf(out_file,"%f\t",tout4[j]);
					fprintf(out_file,"%f\t",stateout4[0][j]);
					fprintf(out_file,"%f\n",stateout4[1][j]);
					}
				eventdef.direction[0]=-eventdef.direction[0];
				printf("event %d detected, data saved in out4.dat\n",i);					
			}
			sysdef.t0=tout4[nelout4-1];
			sysdef.x0[0]=stateout4[0][nelout4-1];
			sysdef.x0[1]=stateout4[1][nelout4-1];
			free(tout4);
			for(j=0;j<2;j++)
				free(stateout4[j]);
			free(stateout4);
		}
		else printf("flag = %d\n", flag);
	}
	fclose(out_file);
	
	return 0;
}


/*----------------------------------------------------------------------------------------*/
int rhsfun(const double time, const double state[2], double rhs[2], void *pars){
	double a,b;
	a=((double*)pars)[0];
	b=((double*)pars)[1];
	rhs[0]=a*state[0]+b*state[1];
	rhs[1]=-b*state[0]+a*state[1];
	return 0;
}
/*----------------------------------------------------------------------------------------*/
double eventfun(const double t, const double x[2], const int event_index,  const void *pars){
	switch(event_index){
		case 0: return x[1];
		default:{
			fprintf(stderr,"Error, event number %d undefined\n",event_index);
			exit(-1);
		}
	}
}
/*----------------------------------------------------------------------------------------*/
